from django.views.generic import RedirectView


class NodeRedirectView(RedirectView):
    """Redirect to the target Node."""
    def get_redirect_url(self, *args, **kwargs):
        redirect = kwargs['node']
        self.permanent = redirect.permanent
        return redirect.target.url



from django.views.generic import DetailView


class PageDetail(DetailView):
    """
    Show a Page in all its glory!

    Requires `handler` in the kwargs in order to get the Page instance.
    """
    def get_object(self):
        return self.kwargs['node']



from django.test import TestCase

from ..models import NodeRedirect
from .factories import ChildNodeRedirectFactory
from conman.nav_tree.tests.factories import RootNodeFactory
from conman.nav_tree.tests.test_models import NODE_BASE_FIELDS


class NodeRedirectTest(TestCase):
    def test_fields(self):
        expected = (
            'id',
            'node_ptr',
            'node_ptr_id',
            'target',
            'target_id',
            'permanent',
        ) + NODE_BASE_FIELDS
        fields = NodeRedirect._meta.get_all_field_names()
        self.assertCountEqual(fields, expected)


class NodeRedirectUnicodeMethodTest(TestCase):
    """We should get something nice when RedirectNode is cast to string"""
    def test_str(self):
        leaf = ChildNodeRedirectFactory.create(slug='leaf')

        self.assertEqual(str(leaf), 'NodeRedirect @ /leaf/')



from incuna_test_utils.testcases.integration import BaseIntegrationTestCase
from incuna_test_utils.testcases.request import BaseRequestTestCase

from .factories import UserFactory


class RequestTestCase(BaseRequestTestCase):
    user_factory = UserFactory


class IntegrationTestCase(BaseIntegrationTestCase):
    user_factory = UserFactory



from django.core.urlresolvers import resolve, Resolver404
from django.test import TestCase

from .. import views


class TestNavTreeURLRouter(TestCase):
    def assert_url_uses_router(self, url):
        resolved_view = resolve(url)
        expected_view = views.node_router
        self.assertEqual(resolved_view.func, expected_view)

    def test_blank_url(self):
        """Blank urls should not resolve.

        This is actually a test of django, as urls must start with `/`.
        """
        with self.assertRaises(Resolver404):
            self.assert_url_uses_router('')

    def test_double_slash_url(self):
        """Trailing slashes should trail something."""
        with self.assertRaises(Resolver404):
            self.assert_url_uses_router('//')

    def test_root_url(self):
        self.assert_url_uses_router('/')

    def test_child_url(self):
        self.assert_url_uses_router('/slug/')

    def test_nested_child_url(self):
        self.assert_url_uses_router('/foo/bar/')

    def test_numerical_url(self):
        self.assert_url_uses_router('/meanings/42/')

    def test_without_trailing_slash(self):
        with self.assertRaises(Resolver404):
            self.assert_url_uses_router('/fail')



from django.conf.urls import url


def dummy_view():
    """Always gets mocked."""


view_path = 'conman.nav_tree.tests.urls.dummy_view'


urlpatterns = [
    url(r'^$', view_path),
    url(r'^(?P<slug>[a-zA-Z0-9_-]+)/$', view_path),
]



from unittest import mock

from django.http import HttpResponse
from django.test import TestCase

from .. import views
from . import factories


class RouterTest(TestCase):
    """Test that `node_router` correctly deals with the urls handed to it."""
    def test_root(self):
        url = ''
        factories.NodeFactory.create()
        request = mock.MagicMock()
        handle_path = 'conman.nav_tree.models.Node.handle'
        with mock.patch(handle_path) as handle:
            response = views.node_router(request, url)

        handle.assert_called_with(request, '/' + url)
        self.assertEqual(response, handle(request, '/' + url))

    def test_complex_url(self):
        url = 'slug/42/foo/bar/'
        factories.NodeFactory.create()
        request = mock.MagicMock()
        handle_path = 'conman.nav_tree.models.Node.handle'
        with mock.patch(handle_path) as handle:
            response = views.node_router(request, url)

        handle.assert_called_with(request, '/' + url)
        self.assertEqual(response, handle(request, '/' + url))


class RouterIntegrationTest(TestCase):
    """Test that `node_router` is correctly handed urls."""
    def test_root_url(self):
        url = '/'
        factories.NodeFactory.create()
        handle_path = 'conman.nav_tree.models.Node.handle'
        with mock.patch(handle_path) as handle:
            handle.return_value = HttpResponse()
            response = self.client.get(url)

        handle.assert_called_with(response.wsgi_request, url)

    def test_complex_url(self):
        url = '/slug/42/foo/bar/'
        factories.NodeFactory.create()
        handle_path = 'conman.nav_tree.models.Node.handle'
        with mock.patch(handle_path) as handle:
            handle.return_value = HttpResponse()
            response = self.client.get(url)

        handle.assert_called_with(response.wsgi_request, url)



#! /usr/bin/env python
"""From http://stackoverflow.com/a/12260597/400691"""
import sys

from colour_runner.django_runner import ColourRunnerMixin
import dj_database_url
import django
from django.conf import settings
from django.test.runner import DiscoverRunner


settings.configure(
    DATABASES={'default': dj_database_url.config(
        default='postgres://localhost/conman',
    )},
    DEFAULT_FILE_STORAGE='inmemorystorage.InMemoryStorage',
    INSTALLED_APPS=(
        # Put contenttypes before auth to work around test issue.
        # See: https://code.djangoproject.com/ticket/10827#comment:12
        'conman.nav_tree',
        'conman.pages',
        'conman.redirects',

        'mptt',
        'polymorphic',
        'polymorphic_tree',

        'django.contrib.admin',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.sites',
    ),
    PASSWORD_HASHERS=('django.contrib.auth.hashers.MD5PasswordHasher',),
    SITE_ID=1,
    ROOT_URLCONF='conman.tests.urls',
    MIDDLEWARE_CLASSES=(),
    NAV_NODE_HANDLERS=(),
)


django.setup()


class TestRunner(ColourRunnerMixin, DiscoverRunner):
    pass


test_runner = TestRunner(verbosity=1)
failures = test_runner.run_tests(['conman'])
if failures:
    sys.exit(1)



from django.db import models

from conman.nav_tree.models import Node


class NodeRedirect(Node):
    """
    When `node` is browsed to, browser should be redirected to `target`.

    This model holds the data required to make that connection.
    """
    target = models.ForeignKey('nav_tree.Node', related_name='+')
    permanent = models.BooleanField(default=False, blank=True)



from django.test import TestCase

from ..handlers import NodeRedirectHandler
from ..views import NodeRedirectView
from conman.nav_tree.handlers import SimpleHandler


class TestNodeRedirectHandler(TestCase):
    def test_heritage(self):
        self.assertTrue(issubclass(NodeRedirectHandler, SimpleHandler))

    def test_view(self):
        view = NodeRedirectHandler.view
        expected = NodeRedirectView.as_view()

        self.assertEqual(view.__name__, expected.__name__)
        self.assertEqual(view.__module__, expected.__module__)



import factory

from .. import models


class NodeFactory(factory.DjangoModelFactory):
    class Meta:
        model = models.Node


class RootNodeFactory(NodeFactory):
    slug = ''
    parent = None

    class Meta:
        django_get_or_create = ['slug']


class ChildNodeFactory(NodeFactory):
    parent = factory.SubFactory(RootNodeFactory)
    slug = factory.Sequence('slug{}'.format)



from django.test import TestCase

from .. import views
from . import factories
from conman.tests.utils import RequestTestCase


class TestPageDetail(RequestTestCase):
    def test_get_object(self):
        request = self.create_request()
        page = factories.PageFactory.create()

        view = views.PageDetail(request=request, kwargs={'node': page})
        obj = view.get_object()

        self.assertEqual(obj, page)


class TestPageDetailIntegration(TestCase):
    view = views.PageDetail

    def test_get(self):
        page = factories.PageFactory.create(content='This is a test')
        response = self.client.get(page.url)
        self.assertIn(page.content, response.rendered_content)



def simple_node_view(request, node):
    """
    Delgate handling this request to the view in handler's attribute, `view`.

    It is worth noting that `view` is not a bound method of `handler`, and so
    it is required to pass `handler` to it.
    """
    return node.get_handler().view(request, node=node)



import factory

from .. import handlers, models
from conman.nav_tree.tests.factories import ChildNodeFactory, RootNodeFactory


class NodeRedirectFactoryMixin(factory.Factory):
    handler = handlers.NodeRedirectHandler.path()

    class Meta:
        model = models.NodeRedirect


class RootNodeRedirectFactory(NodeRedirectFactoryMixin, RootNodeFactory):
    pass


class ChildNodeRedirectFactory(NodeRedirectFactoryMixin, ChildNodeFactory):
    target = factory.SubFactory(ChildNodeFactory)



from django.test import TestCase

from .. import views
from .factories import ChildNodeRedirectFactory
from conman.tests.utils import RequestTestCase
from conman.nav_tree.tests.factories import ChildNodeFactory


class TestNodeRedirectView(RequestTestCase):
    def setUp(self):
        self.target = ChildNodeFactory.create()
        self.request = self.create_request()
        self.view = views.NodeRedirectView.as_view()

    def test_permanent(self):
        node = ChildNodeRedirectFactory.create(
            target=self.target,
            permanent=True,
        )
        response = self.view(self.request, node=node)

        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'], self.target.url)

    def test_temporary(self):
        node = ChildNodeRedirectFactory.create(
            target=self.target,
            permanent=False,
        )
        response = self.view(self.request, node=node)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], self.target.url)


class TestNodeRedirectViewIntegration(TestCase):
    def setUp(self):
        self.target = ChildNodeFactory.create()
        self.expected = 'http://testserver' + self.target.url

    def test_permanent(self):
        node = ChildNodeRedirectFactory.create(
            target=self.target,
            permanent=True,
        )
        response = self.client.get(node.url)

        self.assertEqual(response.status_code, 301)
        self.assertEqual(response['Location'], self.expected)

    def test_temporary(self):
        node = ChildNodeRedirectFactory.create(
            target=self.target,
            permanent=False,
        )
        response = self.client.get(node.url)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], self.expected)



from .models import Node


def node_router(request, url):
    """Catch-all view that delegates view handling to the best Node match."""
    # Django strips the leading / when resolving urls, so we'll just go ahead
    # and add it again. This allows us to use it for resolving later.
    url = '/' + url
    node = Node.objects.best_match_for_path(url)
    return node.handle(request, url)



from ..base import BaseHandler


class UnboundViewMeta(type):
    """
    Metaclass that wraps the `view` attribute with `staticmethod`.

    This ensures that the method does not bind to the class unintentionally.
    """
    def __new__(cls, name, bases, attrs):
        view = attrs.get('view')
        if view:
            attrs['view'] = staticmethod(view)
        return super().__new__(cls, name, bases, attrs)


class SimpleHandler(BaseHandler, metaclass=UnboundViewMeta):
    """
    Abstract handler for Nodes that have one url: `/` relative to the Node.

    Subclasses should define a view on the class as `view`. This will be
    called if the `path` passed to `handle` is `/`.
    """
    urlconf = 'conman.nav_tree.handlers.simple.urls'



import importlib
import os


def split_path(path):
    paths = ['/']
    path = path.rstrip('/')

    while path:
        paths.append(path + '/')
        path = os.path.split(path)[0]
        if path == '/':
            break
    return paths


def import_from_dotted_path(path):
    """
    Imports an object (class/module/etc) from a python path string.

    The path must have at least one dot.
    """
    module_path, attr = path.rsplit('.', 1)
    module = importlib.import_module(module_path)
    return getattr(module, attr)



from django.test import TestCase

from .. import handlers, views
from conman.nav_tree.handlers import SimpleHandler


class TestPageHandler(TestCase):
    def test_heritage(self):
        self.assertTrue(issubclass(handlers.PageHandler, SimpleHandler))

    def test_view(self):
        view = handlers.PageHandler.view
        expected = views.PageDetail.as_view()

        self.assertEqual(view.__name__, expected.__name__)
        self.assertEqual(view.__module__, expected.__module__)



import inspect

from django.conf import settings
from django.core import checks
from django.db import models
from polymorphic_tree.managers import PolymorphicMPTTModelManager
from polymorphic_tree.models import (
    PolymorphicMPTTModel,
    PolymorphicTreeForeignKey,
)

from .utils import import_from_dotted_path, split_path


class NodeManager(PolymorphicMPTTModelManager):
    def best_match_for_path(self, path):
        """
        Return the best match for a path. If the path as given is unavailable,
        continues to search by chopping path components off the end.

        Tries hard to avoid unnecessary database lookups by comparing all
        possible matching URL prefixes and choosing the longest match.

        Node.objects.best_match_for_path('/photos/album/2008/09') might return
        the Node with url '/photos/album/'.

        Adapted from feincms/module/page/models.py:71 in FeinCMS v1.9.5.
        """
        paths = split_path(path)

        extra = {'length': 'LENGTH(url)'}
        qs = self.filter(url__in=paths).extra(select=extra).order_by('-length')
        try:
            return qs[0]
        except IndexError:
            msg = 'No matching Node for URL. (Have you made a root Node?)'
            raise self.model.DoesNotExist(msg)


class Node(PolymorphicMPTTModel):
    HANDLER_CHOICES = settings.NAV_NODE_HANDLERS

    parent = PolymorphicTreeForeignKey('self', null=True, blank=True, related_name='children')
    handler = models.CharField(max_length=255, choices=HANDLER_CHOICES)
    slug = models.SlugField(max_length=255, default='', help_text='''
        Used to create the location of the Node. The Root Node needs
        "slug" to be blank; all other Nodes need a value unique to the parent.
        It can only contain letters, numbers, underscores, or hyphens.
    ''')
    # Cached location in tree. Reflects parent and slug on self and ancestors.
    url = models.TextField(db_index=True, editable=False, unique=True)

    objects = NodeManager()

    class Meta:
        unique_together = ('parent', 'slug')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset_originals()

    def __str__(self):
        return '{} @ {}'.format(self.__class__.__name__, self.url)

    def get_handler_class(self):
        """Imports a class from the python path string in `self.handler`."""
        return import_from_dotted_path(self.handler)

    def get_handler(self):
        """
        Get an instance of the handler for this Node instance.

        Multiple calls to this method (on the same instance of Node) will
        return the same instance of handler.
        """
        try:
            return self._handler
        except AttributeError:
            self._handler = self.get_handler_class()(self)
            return self._handler

    def handle(self, request, path):
        """
        Delegate handling the request to the handler.

        The path of this node is chopped off the url to save the handler from
        needing to deal with it. If it really needs it, it will be able to
        derive it from the node (self) that is passed to it on instantiation.
        """
        handler = self.get_handler()
        # Strip the node url from the rest of the path
        path = path[len(self.url) - 1:]
        # Deal with the request
        return handler.handle(request, path)

    def reset_originals(self):
        """
        Cache a copy of the loaded `url` value.

        This is so we can determine if it has been changed on save.
        """
        self._original_parent_id = self.parent_id
        self._original_slug = self.slug

    def save(self, *args, **kwargs):
        """
        Update the `url` attribute of this node and all descendants.

        Quite expensive when called with a node high up in the tree.

        Adapted from feincms/module/page/models.py:248 in FeinCMS v1.9.5.
        """
        is_root = self.parent_id is None
        has_slug = bool(self.slug)

        # Must be one or the other
        if is_root == has_slug:
            raise ValueError('Node can be a root, or have a slug, not both.')

        def make_url(parent_url, slug):
            return '{}{}/'.format(parent_url, slug)

        parent_changed = self._original_parent_id != self.parent_id
        slug_changed = self._original_slug != self.slug
        url_changed = parent_changed or slug_changed or not self.url

        if url_changed:
            self.url = '/' if is_root else make_url(self.parent.url, self.slug)

        super().save(*args, **kwargs)
        self.reset_originals()

        # If the URL changed we need to update all descendants to
        # reflect the changes. Since this is a very expensive operation
        # on large sites we'll check whether our `url` actually changed
        # or if the updates weren't navigation related:
        if not url_changed:
            return

        nodes = self.get_descendants().order_by('lft')

        cached_urls = {self.id: self.url}
        for node in nodes:
            parent_path = cached_urls[node.parent_id]
            node.url = cached_urls[node.id] = make_url(parent_path, node.slug)

            # Skip this logic on save so we do not recurse.
            super(Node, node).save()
    save.alters_data = True

    @classmethod
    def check(cls, **kwargs):
        """
        Check that the requirements for this class are correctly set up.

        In particular, this checks that each path in NAV_NODE_HANDLERS can be
        imported, and that the resultant object is a class.

        I decided against checking that the classes were subclasses of
        `BaseHandler` to allow for ducktyping, but perhaps in future this could
        check that the class has a "handle" method, and can take a Node
        instance as an arg on instantiation.
        """
        errors = super().check(**kwargs)
        for path, name in settings.NAV_NODE_HANDLERS:
            try:
                imported = import_from_dotted_path(path)
            except (ImportError, ValueError):
                msg = "Error importing '{}' from NAV_NODE_HANDLERS"
                errors.append(checks.Error(
                    msg.format(path),
                    hint='This setting must be a dotted python path',
                    obj=cls,
                ))
                continue

            if not inspect.isclass(imported):
                msg = "Expected '{}' from NAV_NODE_HANDLERS to be a class"
                hint = 'This setting must reference a class, not a {}'
                errors.append(checks.Error(
                    msg.format(path),
                    hint=hint.format(type(imported)),
                    obj=cls,
                ))

        return errors



from django.core.urlresolvers import resolve


class BaseHandler:
    """
    Abstract base class for `Node` handlers.

    Subclasses should define a `urlconf` property as a dotted path. This will
    be used to resolve a view when handling requests.

    Views referenced in the `urlconf` will receive `node` as a kwarg, as
    well as the other args and kwargs they would expect given their urlpattern.
    """
    @classmethod
    def path(cls):
        """Get dotted-path of this class"""
        return '{}.{}'.format(cls.__module__, cls.__name__)

    def __init__(self, node):
        """Store the Node so that we know what we're handling."""
        self.node = node

    def handle(self, request, path):
        """
        Resolve `path` to a view, and get it to handle the `request`.

        Returns the `HttpResponse` returned by the view.

        Raises `django.core.urlresolvers.Resolver404` if `path` isn't found.
        """
        view, args, kwargs = resolve(path, urlconf=self.urlconf)
        return view(request, *args, node=self.node, **kwargs)



from sirtrevor.fields import SirTrevorField

from conman.nav_tree.models import Node


class Page(Node):
    """A basic Page of content provided by Sir Trevor"""
    content = SirTrevorField(default='')



from .. import handlers, models
from conman.nav_tree.tests.factories import RootNodeFactory


class PageFactory(RootNodeFactory):
    handler = handlers.PageHandler.path()

    class Meta:
        model = models.Page



from django.test import TestCase

from .. import models
from conman.nav_tree.tests.test_models import NODE_BASE_FIELDS


class PageTest(TestCase):
    def test_fields(self):
        expected = (
            'id',
            'node_ptr',
            'node_ptr_id',
            'content',
        ) + NODE_BASE_FIELDS
        fields = models.Page._meta.get_all_field_names()
        self.assertCountEqual(fields, expected)



from unittest import mock

from django.core.urlresolvers import clear_url_caches, Resolver404
from django.test import TestCase
from django.views.generic import View

from ..handlers import BaseHandler, SimpleHandler


class BaseHandlerPathTest(TestCase):
    """Test BaseHandler.path()"""
    def test_path(self):
        """Test directly on base class"""
        base_handler_path = 'conman.nav_tree.handlers.base.BaseHandler'
        self.assertEqual(BaseHandler.path(), base_handler_path)

    def test_path_on_subclass(self):
        """Test on subclass"""
        class TestHandler(BaseHandler):
            __module__ = 'does_not_exist'

        self.assertEqual(TestHandler.path(), 'does_not_exist.TestHandler')


class BaseHandlerInitTest(TestCase):
    """Test BaseHandler.__init__()"""
    def test_init(self):
        """Make sure "node" gets saved on the handler"""
        node = mock.MagicMock()

        handler = BaseHandler(node)

        self.assertEqual(handler.node, node)


class BaseHandlerHandleTest(TestCase):
    """Test BaseHandler.handle()"""
    def setUp(self):
        class TestHandler(BaseHandler):
            urlconf = 'conman.nav_tree.tests.urls'

        self.node = mock.Mock()
        self.request = mock.Mock()
        self.handler = TestHandler(self.node)
        self.view = 'conman.nav_tree.tests.urls.dummy_view'

    def tearDown(self):
        """Stops the tests leaking into each other through the url cache"""
        clear_url_caches()

    def test_handle_basic(self):
        """Show that url resolving works at the root of the urlconf"""
        with mock.patch(self.view) as view:
            response = self.handler.handle(self.request, '/')

        view.assert_called_with(self.request, node=self.node)
        self.assertEqual(response, view(self.request, node=self.node))

    def test_handle_slug(self):
        """Show that url resolving works with slugs"""
        slug = 'slug'
        with mock.patch(self.view) as view:
            response = self.handler.handle(self.request, '/slug/')

        view.assert_called_with(self.request, node=self.node, slug=slug)

        expected = view(self.request, node=self.node, slug=slug)
        self.assertEqual(response, expected)

    def test_handle_no_url_match(self):
        """Show that an error is thrown when the url does not match"""
        with self.assertRaises(Resolver404):
            with mock.patch(self.view) as view:
                self.handler.handle(self.request, '/no/match/')

        self.assertFalse(view.called)


class SimpleHandlerHandleTest(TestCase):
    """Test SimpleHandler.handle()"""
    def setUp(self):
        class TestHandler(SimpleHandler):
            view = mock.MagicMock()

        self.node = mock.Mock()
        self.request = mock.Mock()
        self.handler = TestHandler(self.node)
        self.node.get_handler.return_value = self.handler
        self.view = TestHandler.view

    def tearDown(self):
        """Stops the tests leaking into each other through the url cache"""
        clear_url_caches()

    def test_handle_basic(self):
        """Show that SimpleHandler.view is used to process the request"""
        response = self.handler.handle(self.request, '/')

        self.view.assert_called_with(self.request, node=self.node)
        expected = self.view(self.request, node=self.node)
        self.assertEqual(response, expected)

    def test_handle_slug(self):
        """Show that slugs are not accepted"""
        with self.assertRaises(Resolver404):
            self.handler.handle(self.request, '/slug/')

        self.assertFalse(self.view.called)

    def test_handle_pk(self):
        """Show that pks are not accepted"""
        with self.assertRaises(Resolver404):
            self.handler.handle(self.request, '/42/')

        self.assertFalse(self.view.called)


class SimpleHandlerViewBindingTest(TestCase):
    """Views should not unexpectedly "bind" to SimpleHandler subclasses"""
    def test_unbound_function(self):
        """Make sure that the handler is not bound to self on the view"""
        def unbound_function(request, **kwargs):
            return request

        class TestHandler(SimpleHandler):
            view = unbound_function

        handler = TestHandler(None)  # First arg here not used
        request = mock.Mock()
        response = handler.view(request)
        self.assertIs(response, request)

    def test_bound_function(self):
        """Make sure that class based views get the expected args"""
        class TestView(View):
            def dispatch(self, request, node=None):
                return self, request

        class TestHandler(SimpleHandler):
            view = TestView.as_view()

        handler = TestHandler(None)  # First arg here not used
        request = mock.Mock()
        self_arg, request_arg = handler.view(request)
        self.assertIsInstance(self_arg, TestView)
        self.assertIs(request_arg, request)



from setuptools import setup, find_packages


version = '0.0.1a1'

setup(
    name='django-conman',
    packages=find_packages(),
    include_package_data=True,
    version=version,
    description='A modular CMS for django, sponsored by Incuna',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    author='Charlie Denton',
    author_email='charlie@meshy.co.uk',
    url='https://github.com/meshy/django-conman/',
    install_requires=[
        'django-mptt>=0.6.1,<=0.7',
        'django-polymorphic-tree>=1.0b1',
        'django-sirtrevor>=0.2.3,<0.3',
    ],
)



from unittest import mock

from django.db.utils import IntegrityError
from django.test import TestCase

from .. import handlers
from ..models import Node
from .factories import ChildNodeFactory, NodeFactory, RootNodeFactory


NODE_BASE_FIELDS = (
    'handler',
    'parent',
    'parent_id',
    'slug',
    'url',

    # MPTT fields
    'level',
    'lft',
    'rght',
    'tree_id',

    # Polymorphic fields
    'polymorphic_ctype',
    'polymorphic_ctype_id',

    # Incoming foreign keys
    'children',  # FK from self. The other end of "parent".
)


class NodeTest(TestCase):
    def test_fields(self):
        expected = (
            'id',
            'noderedirect',
            'page',
        ) + NODE_BASE_FIELDS
        fields = Node._meta.get_all_field_names()
        self.assertCountEqual(fields, expected)


class NodeValidateOnSave(TestCase):
    def test_create_root_with_slug(self):
        """Root must not have a slug"""
        root_node = NodeFactory.build(slug='slug', parent=None)

        with self.assertRaises(ValueError):
            root_node.save()

    def test_create_leaf_without_slug(self):
        """Leaf nodes must have a slug"""
        root_node = RootNodeFactory.create()
        leaf = NodeFactory.build(slug='', parent=root_node)

        with self.assertRaises(ValueError):
            leaf.save()


class NodeUniqueness(TestCase):
    def test_unique_slug_per_parent(self):
        slug = 'slug'
        root_node = RootNodeFactory.create()
        NodeFactory.create(slug=slug, parent=root_node)

        with self.assertRaises(IntegrityError):
            NodeFactory.create(slug=slug, parent=root_node)

    def test_unique_root_url(self):
        Node.objects.create(slug='')

        with self.assertRaises(IntegrityError):
            Node.objects.create(slug='')


class NodeSkipUpdateWithoutChange(TestCase):
    def test_no_update_without_changes(self):
        """Saving unchanged Node shouldn't query parent to rebuild the url."""
        branch = ChildNodeFactory.create(slug='branch')
        branch = Node.objects.get(pk=branch.pk)
        # Prove that no attempt is made to update descendants.
        with self.assertNumQueries(1):
            # One query:
            # * Update the root.
            branch.save()

    def test_no_update_on_resave(self):
        """Resaving changed Node should only update descendants once."""
        branch = ChildNodeFactory.create(slug='branch')
        NodeFactory.create(slug='leaf', parent=branch)
        branch.slug = 'new_slug'
        branch.save()

        # Prove that no attempt is made to update descendants.
        with self.assertNumQueries(1):
            # One query:
            # * Update the root.
            branch.save()


class NodeCachesURLOnCreateTest(TestCase):
    def setUp(self):
        self.root = RootNodeFactory.create()

    def test_create_root(self):
        """Root node should be at the root url."""
        self.assertEqual(self.root.url, '/')

    def test_create_leaf_on_root(self):
        """Children of the root should be at /<slug>/."""
        leaf = NodeFactory.create(slug='leaf', parent=self.root)

        self.assertEqual(leaf.url, '/leaf/')

    def test_create_child_of_child(self):
        """Children of children should be at /<parent-slug>/<slug>/."""
        branch = NodeFactory.create(slug='branch', parent=self.root)
        leaf = NodeFactory.create(slug='leaf', parent=branch)

        self.assertEqual(leaf.url, '/branch/leaf/')


class NodeCachesURLOnRenameTest(TestCase):
    def test_rename_leaf(self):
        """Changing slug on a leaf should update the cached url."""
        leaf = ChildNodeFactory.create(slug='foo')

        leaf.slug = 'bar'
        leaf.save()

        self.assertEqual(leaf.url, '/bar/')

    def test_rename_branch(self):
        """Changing a branch slug should update the child url."""
        branch = ChildNodeFactory.create(slug='foo')
        leaf = NodeFactory.create(slug='leaf', parent=branch)

        branch.slug = 'bar'
        branch.save()

        leaf = Node.objects.get(pk=leaf.pk)
        self.assertEqual(leaf.url, '/bar/leaf/')

    def test_rename_trunk(self):
        """Changing a trunk slug should update the grandchild url."""
        trunk = ChildNodeFactory.create(slug='foo')
        branch = NodeFactory.create(slug='branch', parent=trunk)
        leaf = NodeFactory.create(slug='leaf', parent=branch)

        trunk.slug = 'bar'
        trunk.save()

        leaf = Node.objects.get(pk=leaf.pk)
        self.assertEqual(leaf.url, '/bar/branch/leaf/')


class NodeCachesURLOnMoveTest(TestCase):
    def test_move_leaf(self):
        """Moving a leaf onto a new branch should update the cached url."""
        branch = ChildNodeFactory.create(slug='foo')
        leaf = NodeFactory.create(slug='leaf', parent=branch)

        new_branch = ChildNodeFactory.create(slug='bar')
        leaf.parent = new_branch
        leaf.save()

        self.assertEqual(leaf.url, '/bar/leaf/')

    def test_move_branch(self):
        """Moving a branch onto a new trunk should update the leaf urls."""
        trunk = ChildNodeFactory.create(slug='foo')
        branch = NodeFactory.create(slug='branch', parent=trunk)
        leaf = NodeFactory.create(slug='leaf', parent=branch)

        new_trunk = ChildNodeFactory.create(slug='bar')
        branch.parent = new_trunk
        branch.save()

        leaf = Node.objects.get(pk=leaf.pk)
        self.assertEqual(leaf.url, '/bar/branch/leaf/')


class NodeManagerBestMatchForPathTest(TestCase):
    """
    Test Node.objects.best_match_for_path works with perfect url matches.

    All of these tests assert use of only one query:
        * Get the best node based on url:
            SELECT
                (LENGTH(url)) AS "length",
                <other fields>
            FROM "nav_tree_node"
            WHERE
                "nav_tree_node"."url" IN (
                    '/',
                    '/url/',
                    '/url/split/',
                    '/url/split/into/',
                    '/url/split/into/bits/')
            ORDER BY "length" DESC
            LIMIT 1
    """
    def test_get_root(self):
        root = RootNodeFactory.create()
        with self.assertNumQueries(1):
            node = Node.objects.best_match_for_path('/')

        self.assertEqual(node, root)

    def test_get_leaf(self):
        leaf = ChildNodeFactory.create(slug='leaf')

        with self.assertNumQueries(1):
            node = Node.objects.best_match_for_path('/leaf/')

        self.assertEqual(node, leaf)

    def test_get_leaf_on_branch(self):
        branch = ChildNodeFactory.create(slug='branch')
        leaf = NodeFactory.create(slug='leaf', parent=branch)

        with self.assertNumQueries(1):
            node = Node.objects.best_match_for_path('/branch/leaf/')

        self.assertEqual(node, leaf)

    def test_get_branch_with_leaf(self):
        branch = ChildNodeFactory.create(slug='branch')
        NodeFactory.create(slug='leaf', parent=branch)

        with self.assertNumQueries(1):
            node = Node.objects.best_match_for_path('/branch/')

        self.assertEqual(node, branch)


class NodeManagerBestMatchForBrokenPathTest(TestCase):
    """
    Test Node.objects.best_match_for_path works without a perfect url match.

    All of these tests assert use of only one query:
        * Get the best node based on url:
            SELECT
                (LENGTH(url)) AS "length",
                <other fields>
            FROM "nav_tree_node"
            WHERE
                "nav_tree_node"."url" IN (
                    '/',
                    '/url/',
                    '/url/split/',
                    '/url/split/into/',
                    '/url/split/into/bits/')
            ORDER BY "length" DESC
            LIMIT 1
    """
    def test_throw_error_without_match(self):
        with self.assertNumQueries(1):
            with self.assertRaises(Node.DoesNotExist):
                Node.objects.best_match_for_path('/')

    def test_fall_back_to_root(self):
        root = RootNodeFactory.create()

        with self.assertNumQueries(1):
            node = Node.objects.best_match_for_path('/absent-branch/')

        self.assertEqual(node, root)

    def test_fall_back_to_branch(self):
        branch = ChildNodeFactory.create(slug='branch')

        with self.assertNumQueries(1):
            node = Node.objects.best_match_for_path('/branch/absent-leaf/')

        self.assertEqual(node, branch)


class NodeGetHandlerClassTest(TestCase):
    def test_get_handler_class(self):
        handler = handlers.BaseHandler
        node = NodeFactory.build(handler=handler.path())

        self.assertEqual(node.get_handler_class(), handler)


class NodeGetHandlerTest(TestCase):
    """Make sure that Node.get_handler acts as expected"""
    def test_get_handler(self):
        """We expect an instance of handler instanciated with a Node"""
        handler_class = handlers.BaseHandler
        node = NodeFactory.build(handler=handler_class.path())

        handler = node.get_handler()
        self.assertIsInstance(handler, handler_class)
        self.assertEqual(handler.node, node)

    def test_get_handler_again(self):
        """Make sure we always get the same instance of a handler on a Node"""
        handler_class = handlers.BaseHandler
        node = NodeFactory.build(handler=handler_class.path())

        first_handler = node.get_handler()
        second_handler = node.get_handler()

        self.assertEqual(first_handler, second_handler)


class NodeHandleTest(TestCase):
    def test_handle(self):
        node = NodeFactory.build(url='/branch/')
        node.get_handler_class = mock.MagicMock()
        request = mock.Mock()

        result = node.handle(request, '/branch/leaf/')

        expected = node.get_handler_class()(node).handle(request, '/leaf/')
        self.assertEqual(result, expected)


class NodeHandlerCheckTest(TestCase):
    """Test the `check` method on Node.

    This method has been written to check that the NAV_NODE_HANDLERS setting
    is valid.

    NAV_NODE_HANDLERS must comply with the format of choices for a CharField,
    but that is already checked by passing it as choices for `Node.handler`.
    """
    def test_no_choices(self):
        """Having no choices should not (yet!) return any errors"""
        with self.settings(NAV_NODE_HANDLERS=[]):
            errors = Node.check()

        self.assertEqual(errors, [])

    def test_good_choices(self):
        """Having ok choices should not return any errors"""
        handlers = [('conman.nav_tree.handlers.BaseHandler', 'Base handler')]

        with self.settings(NAV_NODE_HANDLERS=handlers):
            errors = Node.check()

        self.assertEqual(errors, [])

    def test_wrong_choice_class(self):
        """Having choices that are not classes should return an error"""
        path = 'conman.nav_tree.tests'
        handlers = [(path, 'A module')]

        with self.settings(NAV_NODE_HANDLERS=handlers):
            errors = Node.check()

        self.assertEqual(len(errors), 1)
        expected = "Expected '{}' from NAV_NODE_HANDLERS to be a class"
        self.assertEqual(errors[0].msg, expected.format(path))

    def test_bad_path(self):
        """Having a choice that cannot be imported should return an error"""
        with self.settings(NAV_NODE_HANDLERS=[('broken', 'Invalid path')]):
            errors = Node.check()

        self.assertEqual(len(errors), 1)
        expected = "Error importing 'broken' from NAV_NODE_HANDLERS"
        self.assertEqual(errors[0].msg, expected)


class NodeStrTest(TestCase):
    """Make sure that we get something nice when Node is cast to string"""
    def test_root_str(self):
        """Does it work for a root Node?"""
        node = RootNodeFactory.create()

        self.assertEqual(str(node), 'Node @ /')

    def test_child_str(self):
        """Does it work for a leaf node?"""
        leaf = ChildNodeFactory.create(slug='leaf')

        self.assertEqual(str(leaf), 'Node @ /leaf/')



from django.test import TestCase

from .. import utils


class TestSplitPath(TestCase):
    def test_split_path(self):
        paths = utils.split_path('/a/path/with/many/parts/')
        expected = [
            '/',
            '/a/',
            '/a/path/',
            '/a/path/with/',
            '/a/path/with/many/',
            '/a/path/with/many/parts/',
        ]
        self.assertCountEqual(paths, expected)  # Order does not matter

    def test_split_empty_path(self):
        paths = utils.split_path('')
        expected = ['/']
        self.assertCountEqual(paths, expected)

    def test_split_root_path(self):
        paths = utils.split_path('/')
        expected = ['/']
        self.assertCountEqual(paths, expected)

    def test_split_path_with_dots(self):
        paths = utils.split_path('/path/../')
        expected = [
            '/',
            '/path/',
            '/path/../',
        ]
        self.assertCountEqual(paths, expected)


class TestImportFromDottedPath(TestCase):
    def test_empty(self):
        with self.assertRaises(ValueError):
            utils.import_from_dotted_path('')

    def test_too_short(self):
        with self.assertRaises(ValueError):
            utils.import_from_dotted_path('antigravity')

    def test_import_module(self):
        result = utils.import_from_dotted_path('conman.nav_tree.utils')
        self.assertEqual(result, utils)

    def test_import_class(self):
        this_test = 'conman.nav_tree.tests.test_utils.TestImportFromDottedPath'
        result = utils.import_from_dotted_path(this_test)
        self.assertEqual(result, self.__class__)
